import numpy as np
from numba import njit
from .constants import RE

@njit
def is_visible(sat_pos, ground_ecef, min_elevation):
    sat_vector = sat_pos - ground_ecef
    distance = np.linalg.norm(sat_vector)
    sat_unit_vector = sat_vector / distance
    elevation = np.arcsin(np.dot(sat_unit_vector, ground_ecef) / np.linalg.norm(ground_ecef))
    return elevation >= min_elevation

def calculate_elevation(sat_pos, ground_ecef):
    sat_vector = sat_pos - ground_ecef
    distance = np.linalg.norm(sat_vector)
    sat_unit_vector = sat_vector / distance
    elevation = np.arcsin(np.dot(sat_unit_vector, ground_ecef) / np.linalg.norm(ground_ecef))
    return np.degrees(elevation)

def ground_station_ecef(lat, lon, radius=RE):
    lat_rad = np.radians(lat)
    lon_rad = np.radians(lon)
    x = radius * np.cos(lat_rad) * np.cos(lon_rad)
    y = radius * np.cos(lat_rad) * np.sin(lon_rad)
    z = radius * np.sin(lat_rad)
    return np.array([x, y, z])



import numpy as np
from numba import njit, prange
from .constants import MU, RE, J2

def satellite_positions(times, constellation, backend='cpu', return_frame='ecef', epochs=None, input_type='kepler'):
    """
    Propagate satellites using the selected backend ('cpu' or 'cuda').
    
    Args:
        times: np.ndarray of times (seconds since J2000 or since reference epoch)
        constellation: np.ndarray of shape (N, 6) with either:
                      - Keplerian elements [a, e, inc, Omega, omega, M0] if input_type='kepler'
                      - Position and velocity vectors [rx, ry, rz, vx, vy, vz] if input_type='rv'
                        where positions are in meters and velocities in m/s
        backend: 'cpu' or 'cuda'
        return_frame: Coordinate frame to return ('ecef' or 'eci')
        epochs: np.ndarray of shape (N,) with epoch times in seconds since J2000
                for each satellite. If None, assumes all satellites use times[0] as epoch.
        input_type: 'kepler' for Keplerian elements or 'rv' for position-velocity vectors
        
    Returns:
        np.ndarray of shape (num_sats, num_times, 3) with satellite positions
        in the specified coordinate frame (ECEF by default)
    """
    if backend == 'cpu':
        # Note: CPU backend currently only supports Keplerian elements
        if input_type.lower() != 'kepler':
            raise ValueError(f"CPU backend only supports 'kepler' input_type, not '{input_type}'")
            
        from .prop_cpu import propagate_constellation_cpu
        from .coord_conv import batch_eci_to_ecef, calculate_gmst_from_seconds
        
        # Propagate constellation using CPU with J2 perturbation
        positions_eci = propagate_constellation_cpu(constellation, times, epochs=epochs)
        
        # Transform ECI to ECEF if needed
        if return_frame.lower() == 'ecef':
            # Handle time references for GMST calculation
            if epochs is None:
                # If no epochs provided, use times[0] as reference
                times_seconds = times - times[0]
            else:
                # If epochs provided, use absolute times
                times_seconds = times
                
            # Initialize ECEF positions array
            num_sats = len(constellation)
            num_times = len(times)
            positions_ecef = np.zeros((num_sats, num_times, 3))
            
            # Convert each timestep from ECI to ECEF
            for t in range(num_times):
                seconds = times_seconds[t]
                gmst = calculate_gmst_from_seconds(seconds)
                positions_at_t = positions_eci[:, t, :]
                positions_ecef_at_t = batch_eci_to_ecef(positions_at_t, gmst)
                positions_ecef[:, t, :] = positions_ecef_at_t
                
            return positions_ecef
        else:
            # Return ECI coordinates
            return positions_eci
    elif backend == 'cuda':
        from .prop_cuda import propagate_constellation_cuda
        return propagate_constellation_cuda(constellation, times, return_frame=return_frame, 
                                           epochs=epochs, input_type=input_type)
    else:
        raise ValueError(f"Unknown backend: {backend}")



import numpy as np
from numba import njit, prange
from .constants import MU, RE, J2
import math
@njit(nopython=True, fastmath=True)
def solve_kepler_equation(M, e, tol=1e-10, max_iter=10):
    """
    Solve Kepler's equation for eccentric anomaly using Newton-Raphson method.
    
    Args:
        M: Mean anomaly in radians
        e: Eccentricity
        tol: Convergence tolerance
        max_iter: Maximum number of iterations
        
    Returns:
        E: Eccentric anomaly in radians
    """
    # Initial guess for eccentric anomaly
    E = M
    
    for _ in range(max_iter):
        E_old = E
        f_E = E - e * math.sin(E) - M
        fp_E = 1 - e * math.cos(E)
        
        # Avoid division by zero
        if abs(fp_E) < 1e-10:
            break
            
        delta_E = f_E / fp_E
        E = E - delta_E
        
        # Check for convergence
        if abs(E - E_old) < tol:
            break
            
    return E

@njit(nopython=True, fastmath=True)
def calculate_j2_perturbation_rates(mu, a, e, i):
    """
    Calculate J2 perturbation rates with respect to time (not mean anomaly).
    
    Args:
        mu: Gravitational parameter
        a: Semi-major axis
        e: Eccentricity
        i: Inclination
        
    Returns:
        n0: Unperturbed mean motion
        d_raan_dt: Rate of change of RAAN with respect to time
        d_argper_dt: Rate of change of argument of perigee with respect to time
        d_M_dt: Rate of change of mean anomaly with respect to time
    """
    # Calculate unperturbed mean motion
    n0 = math.sqrt(mu / a**3)
    
    # Calculate perturbation rates per time unit
    p = a * (1 - e**2)
    if p <= 0 or abs(p) < 1e-9:
        # Handle degenerate cases
        return n0, 0.0, 0.0, 0.0
    
    # Calculate J2 perturbation coefficient
    j2_secular_scale = (n0 * RE**2 * J2) / (a**2 * (1 - e**2)**2)
    
    # Calculate rates of change for orbital elements
    d_raan_dt    = -3/2 * j2_secular_scale * math.cos(i)                            # Right ascension of ascending node
    d_argper_dt  =  3/4 * j2_secular_scale * (4 - 5 * math.sin(i)**2)               # Argument of periapsis
    d_M_dt       =  3/4 * j2_secular_scale * math.sqrt(1 - e**2) * (2 - 3 * math.sin(i)**2)  # Mean anomaly
    
    return n0, d_raan_dt, d_argper_dt, d_M_dt

@njit(nopython=True)
def propagate_orbit_j2(mu, a, e, i, raan_0, argper_0, M_0, times):
    """
    Propagate orbit with J2 perturbation for given times.
    
    Args:
        mu: Gravitational parameter
        a: Semi-major axis
        e: Eccentricity
        i: Inclination
        raan_0: Initial right ascension of ascending node
        argper_0: Initial argument of perigee
        M_0: Initial mean anomaly
        times: Array of times since epoch
        
    Returns:
        positions: Array of positions [x, y, z] for each time
    """
    # Calculate unperturbed mean motion and perturbation rates
    n0, d_raan_dt, d_argper_dt, d_M_dt = calculate_j2_perturbation_rates(mu, a, e, i)
    
    positions = np.zeros((len(times), 3))
    
    for idx, t in enumerate(times):
        # Apply perturbations linearly with time (matching CUDA implementation)
        raan_t = raan_0 + d_raan_dt * t
        argper_t = argper_0 + d_argper_dt * t
        M_t = M_0 + (n0 + d_M_dt) * t  
        
        E = solve_kepler_equation(M_t, e)
        
        nu = 2 * np.arctan2(math.sqrt(1 + e) * math.sin(E/2), math.sqrt(1 - e) * math.cos(E/2))
        
        r = a * (1 - e * math.cos(E))
        x_orb = r * math.cos(nu)
        y_orb = r * math.sin(nu)
        
        cos_omega = math.cos(argper_t)
        sin_omega = math.sin(argper_t)
        cos_raan = math.cos(raan_t)
        sin_raan = math.sin(raan_t)
        cos_i = math.cos(i)
        sin_i = math.sin(i)
        
        # Transform to ECI frame
        x = (cos_raan * cos_omega - sin_raan * sin_omega * cos_i) * x_orb + \
            (-cos_raan * sin_omega - sin_raan * cos_omega * cos_i) * y_orb
        y = (sin_raan * cos_omega + cos_raan * sin_omega * cos_i) * x_orb + \
            (-sin_raan * sin_omega + cos_raan * cos_omega * cos_i) * y_orb
        z = (sin_omega * sin_i) * x_orb + (cos_omega * sin_i) * y_orb
        
        positions[idx, 0] = x
        positions[idx, 1] = y
        positions[idx, 2] = z
    
    return positions

@njit(nopython=True, parallel=True)
def propagate_constellation_cpu(satellite_elements, times, epochs=None):
    """
    Propagate entire constellation using CPU acceleration with J2 perturbation.
    
    Args:
        satellite_elements: Array of shape (num_sats, 6) with Keplerian elements
                           [a, e, i, Ω, ω, M] for each satellite
        times: Array of times (seconds since J2000 or since reference epoch)
        epochs: Array of shape (num_sats,) with epoch times in seconds since J2000
               for each satellite. If None, assumes all satellites use times[0] as epoch.
        
    Returns:
        positions: Array of shape (num_sats, num_times, 3) with satellite positions in ECI
    """
    mu = MU
    num_sats = len(satellite_elements)
    num_times = len(times)
    
    # Initialize array for positions
    positions_eci = np.zeros((num_sats, num_times, 3))
    
    # Handle satellite epochs
    if epochs is None:
        # If no epochs provided, use times[0] as reference for all satellites
        times_seconds = times - times[0]
    else:
        # If epochs provided, each satellite has its own reference epoch
        times_seconds = times
    
    # Propagate each satellite in parallel
    for i in prange(num_sats):
        elements = satellite_elements[i]
        a, e, inc, raan, argper, mean_anomaly = elements
        
        # Calculate propagation times relative to this satellite's epoch
        if epochs is not None:
            sat_epoch = epochs[i]
            sat_times = times_seconds - sat_epoch
        else:
            sat_times = times_seconds
        
        # Propagate orbit with J2 perturbation
        positions = propagate_orbit_j2(mu, a, e, inc, raan, argper, mean_anomaly, sat_times)
        positions_eci[i] = positions
    
    return positions_eci



"""
Compare TensorGator's CPU and CUDA implementations to ensure they produce identical results.

"""
import numpy as np
import matplotlib.pyplot as plt
import sys
import os

# Add parent directory to path for TensorGator import
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from tensorgator.propagation import satellite_positions
from tensorgator.constants import MU, RE, J2

def main():
    # --- ISS-like orbit parameters---
    a = RE + 400e3
    e = 0.0
    i = np.radians(51.6)
    raan = 0.0
    argp = 0.0
    M = 0.0

    dt = 300  # 5 minutes
    n_steps = 288  # 24 hours
    times = np.arange(0, dt*n_steps, dt)

    print("TensorGator constants:")
    print(f"  Earth radius (RE): {RE}")
    print(f"  Earth J2: {J2}")
    print(f"  Earth GM (MU): {MU}")
    print("")

    # --- TensorGator CPU Propagation ---
    tg_elements = np.array([[a, e, i, raan, argp, M]])
    tg_cpu_positions = satellite_positions(times, tg_elements, backend='cpu', return_frame='eci')
    tg_cpu_positions = tg_cpu_positions[0]

    # --- TensorGator CUDA Propagation ---
    tg_cuda_positions = satellite_positions(times, tg_elements, backend='cuda', return_frame='eci')
    tg_cuda_positions = tg_cuda_positions[0]

    diff = np.linalg.norm(tg_cpu_positions - tg_cuda_positions, axis=1)
    print(f"Max difference: {np.max(diff):.6f} m, Mean difference: {np.mean(diff):.6f} m")

    import pandas as pd
    df = pd.DataFrame({
        'time_s': times,
        'cpu_x': tg_cpu_positions[:,0],
        'cpu_y': tg_cpu_positions[:,1],
        'cpu_z': tg_cpu_positions[:,2],
        'cuda_x': tg_cuda_positions[:,0],
        'cuda_y': tg_cuda_positions[:,1],
        'cuda_z': tg_cuda_positions[:,2],
        'diff_m': diff
    })
    df.to_csv('cpu_cuda_comparison_intermediates.csv', index=False)

    max_idx = np.argmax(diff)
    print("\nFirst step:")
    print(df.iloc[0])
    print("\nMax difference step:")
    print(df.iloc[max_idx])
    print("\nLast step:")
    print(df.iloc[-1])

    plt.figure(figsize=(8,4))
    plt.plot(times/3600, diff)
    plt.xlabel("Time [hours]")
    plt.ylabel("Position difference [m]")
    plt.title("CPU vs CUDA J2 Propagator Position Difference\n ~1m/day Due to Float32 vs Float64 Precision")
    plt.grid(True)
    plt.tight_layout()
    plt.savefig('cpu_cuda_comparison.png')
    plt.show()

if __name__ == "__main__":
    main()



"""
Coordinate conversion utilities for the HyperGator package.

This module provides functions for converting between different coordinate frames:
- Earth-Centered Inertial (ECI)
- Earth-Centered Earth-Fixed (ECEF)
- Latitude/Longitude/Altitude (LLA)
- Ground station coordinates

It also includes utilities for calculating Greenwich Mean Sidereal Time (GMST).
"""

import numpy as np
import math
from datetime import datetime, timedelta
from numba import njit, vectorize, prange, jit
from .constants import RE, EARTH_ROTATION_RATE

def calculate_gmst(epoch):
    """
    Calculate Greenwich Mean Sidereal Time (GMST) for a given epoch.

    Args:
        epoch: datetime object representing the epoch (UTC/TT)

    Returns:
        GMST in radians, normalized to [0, 2π]
    """
    # 1) Compute Julian centuries since J2000.0
    j2000 = datetime(2000, 1, 1, 12, 0, 0)  # J2000.0 is 2000-01-01 12:00 TT
    delta = epoch - j2000
    days_since_j2000 = delta.total_seconds() / 86400.0
    T = days_since_j2000 / 36525.0

    # 2) GMST at 0h UT (in degrees)
    gmst0 = (
        100.46061837
        + 36000.770053608 * T
        + 0.000387933 * T**2
        - (T**3) / 38710000.0
    )

    # 3) Time-of-day fraction (seconds since last 00:00 UTC)
    midnight = epoch.replace(hour=0, minute=0, second=0, microsecond=0)
    tod_seconds = (epoch - midnight).total_seconds()

    # 4) Add true sidereal rotation since 0h UT:
    #    Earth rotates ~360.98564736629° per mean solar day
    gmst_deg = gmst0 + 360.98564736629 * (tod_seconds / 86400.0)

    gmst_rad = (gmst_deg % 360.0) * (math.pi / 180.0)
    return gmst_rad


from numba import njit

@njit
def calculate_gmst_from_seconds(seconds_since_j2000):
    """
    Calculate GMST from seconds since J2000.0 (2000-01-01 12:00 TT).

    Args:
        seconds_since_j2000: float32/float64 seconds since J2000.0 epoch

    Returns:
        GMST in radians, normalized to [0, 2π]
    """
    # 1) Days and centuries since J2000.0
    days = seconds_since_j2000 / 86400.0
    T = days / 36525.0

    # 2) GMST at 0h UT in degrees
    gmst0 = (
        100.46061837
        + 36000.770053608 * T
        + 0.000387933 * T * T
        - (T**3) / 38710000.0
    )

    # 3) Seconds past last 0h UT
    tod = seconds_since_j2000 % 86400.0

    # 4) Add sidereal rotation
    gmst_deg = gmst0 + 360.98564736629 * (tod / 86400.0)

    # 5) Normalize and convert to radians
    gmst_rad = (gmst_deg % 360.0) * (math.pi / 180.0)
    return gmst_rad

@njit
def eci_to_ecef(position_eci, gmst):
    """
    Convert ECI coordinates to ECEF coordinates using GMST.
    
    Args:
        position_eci: Position vector [x, y, z] in ECI frame
        gmst: Greenwich Mean Sidereal Time in radians
        
    Returns:
        Position vector [x, y, z] in ECEF frame
    """
    x_eci, y_eci, z_eci = position_eci
    
    cos_gmst = math.cos(gmst)
    sin_gmst = math.sin(gmst)
    
    x_ecef = x_eci * cos_gmst + y_eci * sin_gmst
    y_ecef = -x_eci * sin_gmst + y_eci * cos_gmst
    z_ecef = z_eci
    
    return np.array([x_ecef, y_ecef, z_ecef])

@njit
def ecef_to_eci(position_ecef, gmst):
    """
    Convert ECEF coordinates to ECI coordinates using GMST.
    
    Args:
        position_ecef: Position vector [x, y, z] in ECEF frame
        gmst: Greenwich Mean Sidereal Time in radians
        
    Returns:
        Position vector [x, y, z] in ECI frame
    """
    x_ecef, y_ecef, z_ecef = position_ecef
    
    cos_gmst = math.cos(gmst)
    sin_gmst = math.sin(gmst)
    
    x_eci = x_ecef * cos_gmst - y_ecef * sin_gmst
    y_eci = x_ecef * sin_gmst + y_ecef * cos_gmst
    z_eci = z_ecef
    
    return np.array([x_eci, y_eci, z_eci])

def ecef_to_lla(position_ecef):
    """
    Convert ECEF coordinates to latitude, longitude, and altitude.
    
    Args:
        position_ecef: Position vector [x, y, z] in ECEF frame (meters)
        
    Returns:
        Tuple of (latitude, longitude, altitude) in (degrees, degrees, meters)
    """
    x, y, z = position_ecef
    
    # WGS-84 ellipsoid parameters
    a = RE  # Earth's equatorial radius in meters
    f = 1/298.257223563  # Flattening
    b = a * (1 - f)  # Earth's polar radius
    e2 = 1 - (b*b)/(a*a)  # Square of eccentricity
    
    lon = math.atan2(y, x)
    
    p = math.sqrt(x*x + y*y)
    
    # Initial latitude guess
    lat = math.atan2(z, p * (1 - e2))
    
    # Iterative latitude calculation
    for _ in range(5):
        sin_lat = math.sin(lat)
        N = a / math.sqrt(1 - e2 * sin_lat * sin_lat)  
        h = p / math.cos(lat) - N
        lat = math.atan2(z, p * (1 - e2 * N / (N + h)))
    
    lat_deg = math.degrees(lat)
    lon_deg = math.degrees(lon)
    
    sin_lat = math.sin(lat)
    N = a / math.sqrt(1 - e2 * sin_lat * sin_lat)
    h = p / math.cos(lat) - N
    
    return lat_deg, lon_deg, h

def lla_to_ecef(lat, lon, alt):
    """
    Convert latitude, longitude, and altitude to ECEF coordinates.
    
    Args:
        lat: Latitude in degrees
        lon: Longitude in degrees
        alt: Altitude in meters
        
    Returns:
        Position vector [x, y, z] in ECEF frame (meters)
    """
    lat_rad = math.radians(lat)
    lon_rad = math.radians(lon)
    
    # WGS-84 ellipsoid parameters
    a = RE  # Earth's equatorial radius in meters
    f = 1/298.257223563  # Flattening
    e2 = 2*f - f*f  # Square of eccentricity
    
    N = a / math.sqrt(1 - e2 * math.sin(lat_rad) * math.sin(lat_rad))
    
    x = (N + alt) * math.cos(lat_rad) * math.cos(lon_rad)
    y = (N + alt) * math.cos(lat_rad) * math.sin(lon_rad)
    z = (N * (1 - e2) + alt) * math.sin(lat_rad)
    
    return np.array([x, y, z])

def ground_station_ecef(lat, lon, alt=0):
    """
    Calculate ECEF coordinates of a ground station.
    
    Args:
        lat: Latitude in degrees
        lon: Longitude in degrees
        alt: Altitude in meters (default: 0)
        
    Returns:
        Position vector [x, y, z] in ECEF frame (meters)
    """
    return lla_to_ecef(lat, lon, alt)

def cart_to_lat_lon(x, y, z):
    """
    Convert Cartesian ECEF coordinates to latitude and longitude.
    
    Args:
        x, y, z: Cartesian coordinates in ECEF frame (meters)
        
    Returns:
        Tuple of (latitude, longitude) in degrees
    """
    lat, lon, _ = ecef_to_lla(np.array([x, y, z]))
    return lat, lon

@njit(parallel=True)
def batch_eci_to_ecef(positions_eci, gmst):
    """
    Convert batch of ECI coordinates to ECEF coordinates.
    
    Args:
        positions_eci: Array of shape (N, 3) with ECI positions
        gmst: Greenwich Mean Sidereal Time in radians
        
    Returns:
        Array of shape (N, 3) with ECEF positions
    """
    cos_gmst = np.cos(gmst)
    sin_gmst = np.sin(gmst)
    
    n = positions_eci.shape[0]
    positions_ecef = np.zeros_like(positions_eci)
    
    for i in prange(n):
        x_eci = positions_eci[i, 0]
        y_eci = positions_eci[i, 1]
        z_eci = positions_eci[i, 2]
        
        positions_ecef[i, 0] = x_eci * cos_gmst + y_eci * sin_gmst
        positions_ecef[i, 1] = -x_eci * sin_gmst + y_eci * cos_gmst
        positions_ecef[i, 2] = z_eci
    
    return positions_ecef

@njit(parallel=True)
def batch_ecef_to_eci(positions_ecef, gmst):
    """
    Convert batch of ECEF coordinates to ECI coordinates.
    
    Args:
        positions_ecef: Array of shape (N, 3) with ECEF positions
        gmst: Greenwich Mean Sidereal Time in radians
        
    Returns:
        Array of shape (N, 3) with ECI positions
    """
    cos_gmst = np.cos(gmst)
    sin_gmst = np.sin(gmst)
    
    n = positions_ecef.shape[0]
    positions_eci = np.zeros_like(positions_ecef)
    
    for i in prange(n):
        x_ecef = positions_ecef[i, 0]
        y_ecef = positions_ecef[i, 1]
        z_ecef = positions_ecef[i, 2]
        
        positions_eci[i, 0] = x_ecef * cos_gmst - y_ecef * sin_gmst
        positions_eci[i, 1] = x_ecef * sin_gmst + y_ecef * cos_gmst
        positions_eci[i, 2] = z_ecef
    
    return positions_eci



"""
Compare TensorGator's CUDA implementation to Poliastro's J2 propagator.
WE DO NOT EXPECT EXACT MATCHES DUE TO ANALYTIC VS INTEGRATED APPROACHES.

"""

import unittest
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import pandas as pd

from astropy import units as u
from astropy.time import Time, TimeDelta

from poliastro.twobody import orbit
from poliastro import constants
from poliastro.bodies import Earth
from poliastro.core.perturbations import J2_perturbation
from poliastro.core.elements import rv2coe
from poliastro.core.propagation import func_twobody
from poliastro.ephem import EpochsArray
from poliastro.twobody.propagation import CowellPropagator

import sys
import os
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from tensorgator.propagation import satellite_positions

class TestPoliastroComparison(unittest.TestCase):
    """Test to compare TensorGator propagator with Poliastro."""
    
    def setUp(self):
        """Set up test parameters."""
        # Earth parameters
        self.mu = 3.986004418e14  # Earth's gravitational parameter (m^3/s^2)
        self.re = 6378137.0  # Earth's radius (m)
        self.j2 = 1.08262668e-3  # Earth's J2 coefficient
        
        # Test satellites in different orbit regimes
        # Format: [name, a (m), e, i (rad), RAAN (rad), AoP (rad), M (rad)]
        self.test_satellites = [
            # LEO satellite (similar to ISS)
            ["LEO-1", 6878000, 0.0003, np.radians(51.6), np.radians(45), np.radians(30), np.radians(0)],
            
            # MEO satellite (similar to GPS)
            ["MEO-1", 26560000, 0.0001, np.radians(55), np.radians(120), np.radians(60), np.radians(0)],
            
            # GEO satellite
            ["GEO-1", 42164000, 0.0001, np.radians(0.1), np.radians(75), np.radians(0), np.radians(180)],
            
            # Highly eccentric orbit (similar to Molniya)
            ["HEO-1", 26600000, 0.74, np.radians(63.4), np.radians(270), np.radians(270), np.radians(180)]
        ]
        
        # Convert to numpy array for TensorGator (excluding name)
        self.tg_elements = np.array([sat[1:] for sat in self.test_satellites])
        
        # Simulation time parameters
        self.epoch = '2000-01-01 00:00:00.0'
        self.epoch_time = Time(self.epoch, format='iso', scale='utc')
        self.sim_length = 2 * u.hour  # 2 hours simulation
        self.sim_resolution = 1 * u.min  # 1 minute resolution
        self.num_times = int((self.sim_length / self.sim_resolution).decompose().value) + 1
        # For HyperGator: epochs in seconds since J2000 for each satellite
        self.tg_epochs = np.zeros(len(self.test_satellites))  # All start at J2000
        self.prop_times = TimeDelta(np.linspace(0, self.sim_length.to(u.s).value, num=self.num_times) * u.s)
        
        # Time array for TensorGator (seconds)
        self.times = np.linspace(0, self.sim_length.to(u.s).value, self.num_times)
    
    def test_poliastro_comparison(self):
        """Test to compare TensorGator CUDA propagator with Poliastro."""
        print("\nTensorGator vs Poliastro Comparison Results:")
        
        # Propagate using TensorGator
        tg_positions = satellite_positions(
            self.times,
            self.tg_elements,
            backend='cuda',
            return_frame='eci',
            epochs=self.tg_epochs
        )
        
        # Propagate using Poliastro
        poliastro_positions = np.zeros((len(self.test_satellites), self.num_times, 3))
        from astropy import units as u

        for i, sat in enumerate(self.test_satellites):
            # Extract orbital elements
            name = sat[0]
            a = sat[1] * u.m  # Semi-major axis in meters
            e = float(sat[2]) * u.one  # Eccentricity (dimensionless)
            inc = float(sat[3]) * u.rad  # Inclination in radians
            raan = float(sat[4]) * u.rad  # RAAN in radians
            argp = float(sat[5]) * u.rad  # Argument of perigee in radians
            M0 = float(sat[6]) * u.rad  # Mean anomaly in radians
            
            # Create Poliastro orbit
            sat_orbit = orbit.Orbit.from_classical(
                Earth, 
                a, 
                e, 
                inc, 
                raan,
                argp, 
                M0,  
                epoch=self.epoch_time
            )
            
            def f(t0, state, k):
                du_kep = func_twobody(t0, state, k)
                ax, ay, az = J2_perturbation(
                    t0, state, k, J2=Earth.J2.value, R=Earth.R.to(u.km).value
                )
                du_ad = np.array([0, 0, 0, ax, ay, az])
                return du_kep + du_ad
            
            tofs = TimeDelta(np.linspace(0, self.sim_length.to(u.s).value, num=self.num_times) * u.s)
            
            ephem = sat_orbit.to_ephem(
                EpochsArray(sat_orbit.epoch + tofs, method=CowellPropagator(f=f))
            )
            
            positions_km = np.zeros((self.num_times, 3))
            
            for j, t in enumerate(tofs):
                r, _ = sat_orbit.propagate(t, method=CowellPropagator(f=f)).rv()
                
                positions_km[j, 0] = r[0].to(u.km).value
                positions_km[j, 1] = r[1].to(u.km).value
                positions_km[j, 2] = r[2].to(u.km).value
            
            poliastro_positions[i] = positions_km * 1000
        
        differences = np.zeros((len(self.test_satellites), self.num_times))
        percent_differences = np.zeros((len(self.test_satellites), self.num_times))
        
        max_diff = 0
        max_diff_percent = 0
        max_diff_sat = ""
        max_diff_time = 0
        
        for i, sat_data in enumerate(self.test_satellites):
            for t in range(self.num_times):
                diff = np.linalg.norm(tg_positions[i, t] - poliastro_positions[i, t])
                differences[i, t] = diff
                
                radius = np.linalg.norm(poliastro_positions[i, t])
                percent_differences[i, t] = (diff / radius) * 100
                
                if diff > max_diff:
                    max_diff = diff
                    max_diff_percent = percent_differences[i, t]
                    max_diff_sat = sat_data[0]
                    max_diff_time = self.times[t]
        
        print(f"Maximum position difference: {max_diff:.2f} meters")
        print(f"Maximum percentage difference: {max_diff_percent:.6f}%")
        print(f"Occurred for satellite {max_diff_sat} at time {max_diff_time} seconds")
        
        leo_idx = 0
        leo_diff = differences[leo_idx]
        
        final_diff = leo_diff[-1]
        
        sim_duration_days = self.sim_length.to(u.day).value
        diff_per_day = final_diff * (1.0 / sim_duration_days)
        
        print("\nLEO Satellite Analysis:")
        print(f"Simulation duration: {sim_duration_days:.2f} days")
        print(f"Final position difference: {final_diff:.2f} meters")
        print(f"Position difference per day: {diff_per_day:.2f} meters/day")
        print(f"Within expected tolerance (<2km/day): {diff_per_day < 2000}")
        
        intervals = 5
        interval_indices = np.linspace(0, len(self.times)-1, intervals, dtype=int)
        print("\nLEO position differences at intervals:")
        print("Time (hours) | Difference (meters)")
        print("-" * 35)
        for idx in interval_indices:
            time_hours = self.times[idx] / 3600.0
            print(f"{time_hours:10.2f} | {leo_diff[idx]:10.2f}")
        
        if leo_diff[-1] > leo_diff[0]:
            print("\nDivergence is increasing over time")
            from scipy.stats import linregress
            slope, intercept, r_value, p_value, std_err = linregress(self.times / 86400.0, leo_diff)  # time in days
            print(f"Rate of divergence: {slope:.2f} meters/day (R² = {r_value**2:.4f})")
        else:
            print("\nDivergence is not increasing over time")
        
        plt.figure(figsize=(15, 12))
        
        plt.subplot(3, 1, 1)
        for i, sat_data in enumerate(self.test_satellites):
            plt.plot(self.times / 60, differences[i], label=sat_data[0], linewidth=2)
        
        plt.xlabel('Time (minutes)')
        plt.ylabel('Position Difference (meters)')
        plt.title('Absolute Position Differences Between HyperGator and Poliastro')
        plt.grid(True)
        plt.legend()
        
        plt.subplot(3, 1, 2)
        for i, sat_data in enumerate(self.test_satellites):
            plt.plot(self.times / 60, percent_differences[i], label=sat_data[0], linewidth=2)
        
        plt.xlabel('Time (minutes)')
        plt.ylabel('Position Difference (%)')
        plt.title('Percentage Position Differences Relative to Orbit Radius')
        plt.grid(True)
        plt.legend()
        
        ax = plt.subplot(3, 1, 3, projection='3d')
        
        heo_idx = 3  # Index of the HEO satellite
        
        heo_tg = tg_positions[heo_idx]
        heo_poliastro = poliastro_positions[heo_idx]
        
        divergence = heo_tg - heo_poliastro
        
        scale_factor = 1.0
        if np.max(np.abs(divergence)) > 0:
            scale_factor = 1000.0 / np.max(np.abs(divergence))
        
        ax.plot(heo_poliastro[:, 0] / 1000, heo_poliastro[:, 1] / 1000, heo_poliastro[:, 2] / 1000, 'b-', label='Poliastro Path', linewidth=2)
        
        step = max(1, self.num_times // 10)  # Show 10 vectors along the path
        for t in range(0, self.num_times, step):
            ax.quiver(heo_poliastro[t, 0] / 1000, heo_poliastro[t, 1] / 1000, heo_poliastro[t, 2] / 1000,
                     divergence[t, 0] * scale_factor / 1000, divergence[t, 1] * scale_factor / 1000, divergence[t, 2] * scale_factor / 1000,
                     color='r', arrow_length_ratio=0.1)
        
        u, v = np.mgrid[0:2*np.pi:20j, 0:np.pi:10j]
        x = (self.re / 1000) * np.cos(u) * np.sin(v)
        y = (self.re / 1000) * np.sin(u) * np.sin(v)
        z = (self.re / 1000) * np.cos(v)
        ax.plot_surface(x, y, z, color='g', alpha=0.1)
        
        ax.set_xlabel('X (km)')
        ax.set_ylabel('Y (km)')
        ax.set_zlabel('Z (km)')
        ax.set_title(f'Position Divergence for {self.test_satellites[heo_idx][0]} (Scaled by {scale_factor:.1f})')
        
        max_range = np.array([heo_poliastro[:, 0].max() - heo_poliastro[:, 0].min(),
                             heo_poliastro[:, 1].max() - heo_poliastro[:, 1].min(),
                             heo_poliastro[:, 2].max() - heo_poliastro[:, 2].min()]).max() / 2.0
        mid_x = (heo_poliastro[:, 0].max() + heo_poliastro[:, 0].min()) * 0.5 / 1000
        mid_y = (heo_poliastro[:, 1].max() + heo_poliastro[:, 1].min()) * 0.5 / 1000
        mid_z = (heo_poliastro[:, 2].max() + heo_poliastro[:, 2].min()) * 0.5 / 1000
        ax.set_xlim(mid_x - max_range / 1000, mid_x + max_range / 1000)
        ax.set_ylim(mid_y - max_range / 1000, mid_y + max_range / 1000)
        ax.set_zlim(mid_z - max_range / 1000, mid_z + max_range / 1000)
        
        plt.tight_layout()
        plt.savefig('poliastro_comparison.png', dpi=300)
        print("Enhanced visualization saved as 'poliastro_comparison.png'")
        
        plt.figure(figsize=(15, 10))
        
        ax = plt.subplot(111, projection='3d')
        
        heo_tg_pos = tg_positions[heo_idx]
        heo_poliastro_pos = poliastro_positions[heo_idx]
        
        ax.plot(heo_poliastro_pos[:, 0] / 1000, heo_poliastro_pos[:, 1] / 1000, heo_poliastro_pos[:, 2] / 1000, 'b-', linewidth=2, label='Poliastro')
        ax.plot(heo_tg_pos[:, 0] / 1000, heo_tg_pos[:, 1] / 1000, heo_tg_pos[:, 2] / 1000, 'r-', linewidth=2, label='HyperGator')
        
        time_markers = np.linspace(0, self.num_times-1, 5).astype(int)
        for t in time_markers:
            ax.scatter(heo_poliastro_pos[t, 0] / 1000, heo_poliastro_pos[t, 1] / 1000, heo_poliastro_pos[t, 2] / 1000, 
                      c='b', s=50, marker='o')
            ax.scatter(heo_tg_pos[t, 0] / 1000, heo_tg_pos[t, 1] / 1000, heo_tg_pos[t, 2] / 1000, 
                      c='r', s=50, marker='^')
            if t == time_markers[0]:
                ax.text(heo_poliastro_pos[t, 0] / 1000, heo_poliastro_pos[t, 1] / 1000, heo_poliastro_pos[t, 2] / 1000, 
                       f'{self.times[t]/60:.0f} min', color='k')
        
        u, v = np.mgrid[0:2*np.pi:20j, 0:np.pi:10j]
        x = (self.re / 1000) * np.cos(u) * np.sin(v)
        y = (self.re / 1000) * np.sin(u) * np.sin(v)
        z = (self.re / 1000) * np.cos(v)
        ax.plot_surface(x, y, z, color='g', alpha=0.2)
        
        ax.set_xlabel('X (km)')
        ax.set_ylabel('Y (km)')
        ax.set_zlabel('Z (km)')
        ax.set_title(f'Highly Eccentric Orbit Comparison: {self.test_satellites[heo_idx][0]}')
        ax.legend()
        
        max_range = np.array([
            max(heo_poliastro_pos[:, 0].max(), heo_tg_pos[:, 0].max()) - min(heo_poliastro_pos[:, 0].min(), heo_tg_pos[:, 0].min()),
            max(heo_poliastro_pos[:, 1].max(), heo_tg_pos[:, 1].max()) - min(heo_poliastro_pos[:, 1].min(), heo_tg_pos[:, 1].min()),
            max(heo_poliastro_pos[:, 2].max(), heo_tg_pos[:, 2].max()) - min(heo_poliastro_pos[:, 2].min(), heo_tg_pos[:, 2].min())
        ]).max() / 2.0
        
        mid_x = (max(heo_poliastro_pos[:, 0].max(), heo_tg_pos[:, 0].max()) + min(heo_poliastro_pos[:, 0].min(), heo_tg_pos[:, 0].min())) * 0.5 / 1000
        mid_y = (max(heo_poliastro_pos[:, 1].max(), heo_tg_pos[:, 1].max()) + min(heo_poliastro_pos[:, 1].min(), heo_tg_pos[:, 1].min())) * 0.5 / 1000
        mid_z = (max(heo_poliastro_pos[:, 2].max(), heo_tg_pos[:, 2].max()) + min(heo_poliastro_pos[:, 2].min(), heo_tg_pos[:, 2].min())) * 0.5 / 1000
        
        ax.set_xlim(mid_x - max_range / 1000, mid_x + max_range / 1000)
        ax.set_ylim(mid_y - max_range / 1000, mid_y + max_range / 1000)
        ax.set_zlim(mid_z - max_range / 1000, mid_z + max_range / 1000)
        
        plt.tight_layout()
        plt.savefig('heo_orbit_comparison.png', dpi=300)
        print("HEO orbit comparison saved as 'heo_orbit_comparison.png'")
        
        plt.figure(figsize=(15, 10))
        
        for i, sat_data in enumerate(self.test_satellites):
            plt.subplot(2, 2, i+1)
            
            sat_name = sat_data[0]
            
            for j, comp in enumerate(['X', 'Y', 'Z']):
                diff_comp = tg_positions[i, :, j] - poliastro_positions[i, :, j]
                plt.plot(self.times / 60, diff_comp, label=f'{comp} Component', linewidth=2)
            
            plt.xlabel('Time (minutes)')
            plt.ylabel('Position Difference (meters)')
            plt.title(f'Position Component Differences for {sat_name}')
            plt.grid(True)
            plt.legend()
        
        plt.tight_layout()
        plt.savefig('poliastro_component_differences.png', dpi=300)
        print("Component differences visualization saved as 'poliastro_component_differences.png'")
        
        results_data = {
            'Time (s)': self.times
        }
        
        for i, sat_data in enumerate(self.test_satellites):
            sat_name = sat_data[0]
            results_data[f'{sat_name}_Diff_X'] = tg_positions[i, :, 0] - poliastro_positions[i, :, 0]
            results_data[f'{sat_name}_Diff_Y'] = tg_positions[i, :, 1] - poliastro_positions[i, :, 1]
            results_data[f'{sat_name}_Diff_Z'] = tg_positions[i, :, 2] - poliastro_positions[i, :, 2]
            results_data[f'{sat_name}_Diff_Total'] = differences[i]
            results_data[f'{sat_name}_Diff_Percent'] = percent_differences[i]
        
        df = pd.DataFrame(results_data)
        df.to_excel('tensorgator_poliastro_comparison.xlsx', index=False)
        print("Detailed comparison data saved to 'tensorgator_poliastro_comparison.xlsx'")

if __name__ == '__main__':
    unittest.main()



import numpy as np
import math
from numba import njit, prange, cuda
from .constants import RE

@njit
def is_visible(sat_pos, ground_ecef, min_elevation):
    """Check if a satellite is visible from a ground station."""
    sat_vector = sat_pos - ground_ecef
    distance = np.linalg.norm(sat_vector)
    sat_unit_vector = sat_vector / distance
    zenith_unit_vector = ground_ecef / np.linalg.norm(ground_ecef)
    elevation = np.arcsin(np.dot(sat_unit_vector, zenith_unit_vector))
    return elevation >= min_elevation

@cuda.jit(device=True)
def is_visible_device(sat_pos, ground_ecef, min_elevation):
    """CUDA device function to check satellite visibility."""
    # Calculate satellite-to-ground vector
    sat_vector_x = sat_pos[0] - ground_ecef[0]
    sat_vector_y = sat_pos[1] - ground_ecef[1]
    sat_vector_z = sat_pos[2] - ground_ecef[2]
    
    # Calculate distance
    distance = math.sqrt(sat_vector_x*sat_vector_x + sat_vector_y*sat_vector_y + sat_vector_z*sat_vector_z)
    
    # Calculate unit vector
    sat_unit_x = sat_vector_x / distance
    sat_unit_y = sat_vector_y / distance
    sat_unit_z = sat_vector_z / distance
    
    # Calculate ground station magnitude
    ground_mag = math.sqrt(ground_ecef[0]*ground_ecef[0] + ground_ecef[1]*ground_ecef[1] + ground_ecef[2]*ground_ecef[2])
    
    # Calculate elevation angle
    dot_product = sat_unit_x*ground_ecef[0] + sat_unit_y*ground_ecef[1] + sat_unit_z*ground_ecef[2]
    elevation = math.asin(dot_product / ground_mag)
    
    return elevation >= min_elevation

@njit(parallel=True)
def batch_visibility_check(satellite_positions, ground_points, min_elevation):
    """
    Check visibility for multiple ground points in parallel using CPU.
    
    Args:
        satellite_positions: Array of shape (num_sats, num_times, 3)
        ground_points: Array of shape (num_points, 3)
        min_elevation: Minimum elevation angle in radians
        
    Returns:
        visibility: Boolean array of shape (num_points, num_times)
    """
    num_points = len(ground_points)
    num_sats = satellite_positions.shape[0]
    num_times = satellite_positions.shape[1]
    visibility = np.zeros((num_points, num_times), dtype=np.bool_)
    
    for p in prange(num_points):
        ground_pos = ground_points[p]
        for t in range(num_times):
            for s in range(num_sats):
                if is_visible(satellite_positions[s, t], ground_pos, min_elevation):
                    visibility[p, t] = True
                    break
    return visibility

@cuda.jit
def visibility_kernel(sat_positions, ground_points, min_elevation, visibility):
    """
    CUDA kernel for calculating visibility between satellites and ground points.
    
    Args:
        sat_positions: Array of shape (num_sats, num_times, 3)
        ground_points: Array of shape (num_points, 3)
        min_elevation: Minimum elevation angle in radians
        visibility: Output array of shape (num_points, num_times)
    """
    # Get 2D grid indices
    p_idx, t_idx = cuda.grid(2)
    
    # Check bounds
    if p_idx < ground_points.shape[0] and t_idx < sat_positions.shape[1]:
        # Check visibility against all satellites at this time
        for s in range(sat_positions.shape[0]):
            if is_visible_device(sat_positions[s, t_idx], ground_points[p_idx], min_elevation):
                visibility[p_idx, t_idx] = 1
                break

def calculate_visibility_cuda(satellite_positions, ground_points, min_elevation):
    """
    Calculate visibility for all ground points using CUDA.
    
    Args:
        satellite_positions: Array of shape (num_sats, num_times, 3)
        ground_points: Array of shape (num_points, 3)
        min_elevation: Minimum elevation angle in radians
        
    Returns:
        visibility: Boolean array of shape (num_points, num_times)
    """
    num_points = len(ground_points)
    num_times = satellite_positions.shape[1]
    
    # Prepare device arrays
    d_sat_positions = cuda.to_device(satellite_positions)
    d_ground_points = cuda.to_device(ground_points)
    d_visibility = cuda.to_device(np.zeros((num_points, num_times), dtype=np.int32))
    
    # Configure CUDA grid
    threads_per_block = (16, 16)
    blocks_per_grid_x = (num_points + threads_per_block[0] - 1) // threads_per_block[0]
    blocks_per_grid_y = (num_times + threads_per_block[1] - 1) // threads_per_block[1]
    blocks_per_grid = (blocks_per_grid_x, blocks_per_grid_y)
    
    # Launch kernel
    visibility_kernel[blocks_per_grid, threads_per_block](
        d_sat_positions, d_ground_points, min_elevation, d_visibility
    )
    
    # Copy results back to host
    visibility = d_visibility.copy_to_host()
    
    return visibility.astype(bool)

def calculate_max_gaps(visibility, time_step):
    """
    Calculate maximum gap durations for each ground point.
    Non-JIT version that works reliably.
    
    Args:
        visibility: Boolean array of shape (num_points, num_times)
        time_step: Time step in seconds
        
    Returns:
        max_gaps: Array of shape (num_points,) with maximum gap duration in seconds
    """
    num_points = visibility.shape[0]
    num_times = visibility.shape[1]
    max_gaps = np.zeros(num_points)
    
    # Use tqdm if available for progress tracking
    try:
        from tqdm import tqdm
        iterator = tqdm(range(num_points), desc="Calculating gaps")
    except ImportError:
        iterator = range(num_points)
    
    for p in iterator:
        vis_timeline = visibility[p]
        
        # Find the longest sequence of False values (gaps)
        current_gap = 0
        max_gap = 0
        
        for t in range(num_times):
            if not vis_timeline[t]:  # Gap (not visible)
                current_gap += 1
            else:  # Visible
                max_gap = max(max_gap, current_gap)
                current_gap = 0
        
        # Check final gap
        max_gap = max(max_gap, current_gap)
        
        # Convert to time
        max_gaps[p] = max_gap * time_step
    
    return max_gaps

def process_in_chunks(satellite_elements, times, ground_points, min_elevation, chunk_size=1000):
    """
    Process visibility in time chunks to reduce memory usage.
    
    Args:
        satellite_elements: Array of shape (num_sats, 6)
        times: Array of shape (num_times,)
        ground_points: Array of shape (num_points, 3)
        min_elevation: Minimum elevation angle in radians
        chunk_size: Number of time steps to process in each chunk
        
    Returns:
        visibility: Boolean array of shape (num_points, num_times)
    """
    from .propagation import satellite_positions
    
    num_points = len(ground_points)
    num_times = len(times)
    visibility = np.zeros((num_points, num_times), dtype=bool)
    
    for t_start in range(0, num_times, chunk_size):
        t_end = min(t_start + chunk_size, num_times)
        times_chunk = times[t_start:t_end]
        
        # Propagate satellites for this time chunk
        positions_chunk = satellite_positions(times_chunk, satellite_elements, backend='cuda')
        
        # Calculate visibility for this chunk
        visibility_chunk = calculate_visibility_cuda(positions_chunk, ground_points, min_elevation)
        
        # Store results
        visibility[:, t_start:t_end] = visibility_chunk
    
    return visibility



MU = 398600936839470.0  # Standard gravitational parameter for Earth (m^3/s^2)
RE = 6378136.3  # Earth's radius (m)
J2 = 1.08262668355315130e-3  # Earth's J2 coefficient
EARTH_ROTATION_RATE = 7.2921150e-5  # rad/s
#d


import numpy as np
import math
from numba import cuda
from .constants import MU, J2, RE
from .coord_conv import batch_eci_to_ecef, calculate_gmst_from_seconds
from numba import config

def propagate_constellation_cuda(satellite_elements, times, return_frame='ecef', epochs=None, input_type='kepler'):
    mu = MU
    j2 = J2
    re = RE
    num_sats = len(satellite_elements)
    num_times = len(times)
    fpp = np.float32 #adjust to fp64 for sub meter accuracy

    if epochs is None:
        epochs = np.zeros(num_sats, dtype=fpp)
        times_seconds = times - times[0]
    else:
        times_seconds = times

    d_elements = cuda.to_device(np.asarray(satellite_elements, dtype=fpp))
    d_times = cuda.to_device(np.asarray(times_seconds, dtype=fpp))
    d_epochs = cuda.to_device(np.asarray(epochs, dtype=fpp))
    d_positions = cuda.device_array((num_sats, num_times, 3), dtype=fpp)

    threadsperblock = (16, 16)
    blockspergrid_x = (num_sats + threadsperblock[0] - 1) // threadsperblock[0]
    blockspergrid_y = (num_times + threadsperblock[1] - 1) // threadsperblock[1]
    simulate_orbit_batch_cuda[(blockspergrid_x, blockspergrid_y), threadsperblock](
        mu, j2, re, d_elements, d_times, d_epochs, d_positions
    )

    positions_eci = d_positions.copy_to_host()

    if return_frame.lower() == 'ecef':
        positions_ecef = np.zeros_like(positions_eci)
        for t in range(num_times):
            gmst = calculate_gmst_from_seconds(times_seconds[t])
            positions_ecef[:, t, :] = batch_eci_to_ecef(positions_eci[:, t, :], gmst)
        return positions_ecef
    else:
        return positions_eci

@cuda.jit(fastmath=True)
def simulate_orbit_batch_cuda(mu, j2, re, elements, times, epochs, positions):
    sat_idx, time_idx = cuda.grid(2)
    num_sats = elements.shape[0]
    num_times = times.shape[0]

    if sat_idx < num_sats and time_idx < num_times:
        a = elements[sat_idx, 0]
        e = elements[sat_idx, 1]
        i = elements[sat_idx, 2]
        raan = elements[sat_idx, 3]
        argp = elements[sat_idx, 4]
        M0 = elements[sat_idx, 5]

        t = times[time_idx] - epochs[sat_idx]

        x, y, z = calculate_position_device(mu, a, e, i, M0, argp, raan, j2, re, t)

        positions[sat_idx, time_idx, 0] = x
        positions[sat_idx, time_idx, 1] = y
        positions[sat_idx, time_idx, 2] = z

@cuda.jit(device=True, fastmath=True)
def calculate_position_device(mu, a, e, i, M, w, raan, j2, re, t):
    n0 = math.sqrt(mu / a**3)
    j2_secular_scale = (n0 * re**2 * j2) / (a**2 * (1 - e**2)**2)

    d_raan_dt = -3/2 * j2_secular_scale * math.cos(i)
    d_argper_dt = 3/4 * j2_secular_scale * (4 - 5 * math.sin(i)**2)
    d_M_dt = 3/4 * j2_secular_scale * math.sqrt(1 - e**2) * (2 - 3 * math.sin(i)**2)

    raan_t = raan + d_raan_dt * t
    w_t = w + d_argper_dt * t
    M_t = M + (n0 + d_M_dt) * t

    # Replace the original Kepler equation solver with the contour method
    E = kepler_equation_contour_device(M_t, e)
    nu = 2 * math.atan2(math.sqrt(1 + e) * math.sin(E / 2), math.sqrt(1 - e) * math.cos(E / 2))
    r = a * (1 - e * math.cos(E))

    x_orb = r * math.cos(nu)
    y_orb = r * math.sin(nu)

    cos_w = math.cos(w_t)
    sin_w = math.sin(w_t)
    cos_raan = math.cos(raan_t)
    sin_raan = math.sin(raan_t)
    cos_i = math.cos(i)
    sin_i = math.sin(i)

    x = (cos_raan * cos_w - sin_raan * sin_w * cos_i) * x_orb + (-cos_raan * sin_w - sin_raan * cos_w * cos_i) * y_orb
    y = (sin_raan * cos_w + cos_raan * sin_w * cos_i) * x_orb + (-sin_raan * sin_w + cos_raan * cos_w * cos_i) * y_orb
    z = (sin_w * sin_i) * x_orb + (cos_w * sin_i) * y_orb

    return x, y, z

#From https://github.com/oliverphilcox/Keplers-Goat-Herd/blob/main/keplers_goat_herd.py
@cuda.jit(device=True, fastmath=True)
def kepler_equation_contour_device(M, e):
    """
    Solve Kepler's equation, E - e sin E = M, via the contour integration method
    
    Args:
        M (float): Mean anomaly
        e (float): Eccentricity
        
    Returns:
        float: Eccentric anomaly E
    """
    # Use only 2 parameters to match the original function signature
    ell = M
    eccentricity = e
    N_it = 10  # Set a default value for N_it
    
    # Normalize ell to [0, 2π] range
    while ell < 0.0:
        ell += 2.0 * math.pi
    while ell > 2.0 * math.pi:
        ell -= 2.0 * math.pi
    
    # Handle edge cases
    if eccentricity < 1e-10:
        return ell
    
    # Define contour radius
    radius = eccentricity / 2.0
    
    # Define contour center
    center = ell - eccentricity / 2.0
    if ell < math.pi:
        center += eccentricity
        
    # Compute sin and cos of center
    sinC = math.sin(center)
    cosC = math.cos(center)
    
    # Initialize output to center
    output = center
    
    # Precompute e*sin(radius) and e*cos(radius)
    esinRadius = eccentricity * math.sin(radius)
    ecosRadius = eccentricity * math.cos(radius)
    
    # Initialize Fourier coefficients
    ft_gx1 = 0.0
    ft_gx2 = 0.0
    
    # j = 0 piece
    zR = center + radius
    tmpsin = sinC * ecosRadius + cosC * esinRadius
    fxR = zR - tmpsin - ell
    
    # Add to arrays with factor of 1/2 since an edge
    if abs(fxR) > 1e-14:  # Avoid division by zero
        ft_gx2 += 0.5 / fxR
        ft_gx1 += 0.5 / fxR
    
    # j = 1 to N_points pieces
    N_points = N_it - 2
    N_fft = (N_it - 1) * 2
    
    for j in range(N_points):
        # Compute sampling points
        freq = 2.0 * math.pi * (j + 1.0) / N_fft
        exp2R = math.cos(freq)
        exp2I = math.sin(freq)
        
        # Compute real and imaginary components
        ecosR = eccentricity * math.cos(radius * exp2R)
        esinR = eccentricity * math.sin(radius * exp2R)
        exp4R = exp2R * exp2R - exp2I * exp2I
        exp4I = 2.0 * exp2R * exp2I
        coshI = math.cosh(radius * exp2I)
        sinhI = math.sinh(radius * exp2I)
        
        # Compute z in real and imaginary parts
        zR = center + radius * exp2R
        zI = radius * exp2I
        
        # Compute components for f(z(x))
        tmpsin = sinC * ecosR + cosC * esinR  # e sin(zR)
        tmpcos = cosC * ecosR - sinC * esinR  # e cos(zR)
        
        fxR = zR - tmpsin * coshI - ell
        fxI = zI - tmpcos * sinhI
        
        # Compute 1/f(z)
        ftmp = fxR * fxR + fxI * fxI
        if ftmp > 1e-14:  # Avoid division by zero
            fxR = fxR / ftmp
            fxI = fxI / ftmp
            
            ft_gx2 += exp4R * fxR + exp4I * fxI
            ft_gx1 += exp2R * fxR + exp2I * fxI
    
    # j = N_it piece
    zR = center - radius
    tmpsin = sinC * ecosRadius - cosC * esinRadius
    fxR = zR - tmpsin - ell
    
    if abs(fxR) > 1e-14:  # Avoid division by zero
        ft_gx2 += 0.5 / fxR
        ft_gx1 += -0.5 / fxR
    
    # Compute and return the solution E(ell,e)
    if abs(ft_gx1) > 1e-14:  # Avoid division by zero
        output += radius * ft_gx2 / ft_gx1
        
    return output


from setuptools import setup, find_packages

setup(
    name='tensorgator',
    version='0.1.1',
    description='Satellite propagation and coverage visualization toolkit',
    author='ApoPeri',
    author_email='ApoPeri@protonmail.com',
    packages=['tensorgator', 'tensorgator.examples'],
    package_dir={'tensorgator': ''},
    package_data={'tensorgator.examples': ['*.ipynb', '*.png']},
    install_requires=[
        'numpy',
        'matplotlib',
        'numba',
    ],
    python_requires='>=3.7',
    include_package_data=True,
    url='https://github.com/ApoPeri/TensorGator',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)
